"""
This is a basic doctest demonstrating that the package and pydra can both be successfully
imported.

>>> import pydra.compose.bidsapp
"""

try:
    from ._version import __version__
except ImportError:
    raise RuntimeError(
        "Pydra package 'pydra-compose-bidsapp' has not been installed, please use "
        "`pip install -e <path-to-repo>` to install development version"
    )

from .builder import define
from .fields import arg, out
from .task import BidsAppTask as Task, BidsAppOutputs as Outputs

__all__ = ["arg", "out", "define", "Task", "Outputs", "__version__"]



import os
import stat
from pathlib import Path
import shutil
import pytest
from fileformats.medimage import NiftiGzX, NiftiGzXBvec
from pydra.compose import bidsapp
from pydra.compose.bidsapp.task import DEFAULT_BIDS_ID
from pydra.utils import asdict
from fileformats.text import Plain as Text
from fileformats.generic import Directory


MOCK_BIDS_APP_NAME = "mockapp"
MOCK_README = "A dummy readme\n" * 100
MOCK_AUTHORS = ["Dumm Y. Author", "Another D. Author"]

BIDS_INPUTS = [
    bidsapp.arg(name="T1w", path="anat/T1w", type=NiftiGzX),
    bidsapp.arg(name="T2w", path="anat/T2w", type=NiftiGzX),
    bidsapp.arg(name="dwi", path="dwi/dwi", type=NiftiGzXBvec),
]
BIDS_OUTPUTS = [
    bidsapp.out(name="whole_dir", type=Directory),  # whole derivative directory
    bidsapp.out(name="a_file", path="file1", type=Text),
    bidsapp.out(name="another_file", path="file2", type=Text),
]


@pytest.mark.xfail(reason="this test will fail until nipype/pydra#822 is merged")
def test_bids_app_docker(
    bids_validator_app_image: str, nifti_sample_dir: Path, work_dir: Path
):

    bids_dir = work_dir / "bids"

    shutil.rmtree(bids_dir, ignore_errors=True)

    TestBids = bidsapp.define(
        bids_validator_app_image,
        inputs=BIDS_INPUTS,
        outputs=BIDS_OUTPUTS,
    )

    task = TestBids(
        **{
            inpt.name: nifti_sample_dir.joinpath(*inpt.path.split("/")).with_suffix(
                inpt.type.ext
            )
            for inpt in BIDS_INPUTS
        }
    )

    result = task(worker="debug")

    for output in BIDS_OUTPUTS:
        assert Path(getattr(result.output, output.name)).exists()


def test_bids_app_naked(
    mock_bids_app_script: str, nifti_sample_dir: Path, work_dir: Path
):

    # Create executable that runs validator then produces some mock output
    # files
    launch_sh = work_dir / "launch.sh"
    # We don't need to run the full validation in this case as it is already tested by test_run_bids_app_docker
    # so we use the simpler test script.
    with open(launch_sh, "w") as f:
        f.write(mock_bids_app_script)

    os.chmod(launch_sh, stat.S_IRWXU)

    TestBids = bidsapp.define(
        launch_sh,
        inputs=BIDS_INPUTS,
        outputs=BIDS_OUTPUTS,
    )

    task = TestBids(
        **{
            i.name: nifti_sample_dir.joinpath(*i.path.split("/")).with_suffix(
                i.type.ext
            )
            for i in BIDS_INPUTS
        },
    )
    outputs = task(worker="debug")

    for output in asdict(outputs).values():
        assert Path(output).exists()



import logging
import typing as ty
import re
from pathlib import Path
import inspect
import docker.errors
from typing import dataclass_transform
from pydra.compose.base import (
    ensure_field_objects,
    build_task_class,
    check_explicit_fields_are_none,
    extract_fields_from_class,
)
from .fields import arg, out
from .task import BidsAppTask as Task
from .task import BidsAppOutputs as Outputs


logger = logging.getLogger("pydra.compose.bidsapp")


@dataclass_transform(
    kw_only_default=True,
    field_specifiers=(arg,),
)
def define(
    wrapped: type | str | None = None,
    /,
    inputs: list[str | arg] | dict[str, arg | type] | None = None,
    outputs: list[str | out] | dict[str, out | type] | type | None = None,
    bases: ty.Sequence[type] = (),
    outputs_bases: ty.Sequence[type] = (),
    auto_attribs: bool = True,
    name: str | None = None,
    xor: ty.Sequence[str | None] | ty.Sequence[ty.Sequence[str | None]] = (),
) -> Task | ty.Callable[[str | type], Task[ty.Any]]:
    """
    Create an interface for a function or a class.

    Parameters
    ----------
    wrapped : type | callable | None
        The executable to run the app (or entrypoint if running inside a container) or
        class to create an interface for.
    inputs : list[str | Arg] | dict[str, Arg | type] | None
        The inputs to the function or class.
    outputs : list[str | base.Out] | dict[str, base.Out | type] | type | None
        The outputs of the function or class.
    image_tag : str
        the tag of the Docker image to use to run the container. If None, the executable
        is assumed to be in the native env.
    auto_attribs : bool
        Whether to use auto_attribs mode when creating the class.
    name: str | None
        The name of the returned class
    xor: Sequence[str | None] | Sequence[Sequence[str | None]], optional
        Names of args that are exclusive mutually exclusive, which must include
        the name of the current field. If this list includes None, then none of the
        fields need to be set.

    Returns
    -------
    Task
        The task class for the Python function
    """

    def make(wrapped: str | type) -> Task:
        if inspect.isclass(wrapped):
            klass = wrapped
            executable = klass.executable
            image_tag = klass.image_tag
            class_name = klass.__name__
            check_explicit_fields_are_none(klass, inputs, outputs)
            parsed_inputs, parsed_outputs = extract_fields_from_class(
                Task,
                Outputs,
                klass,
                arg,
                out,
                auto_attribs,
                skip_fields=["function"],
            )
        else:
            if isinstance(wrapped, Path):
                wrapped = str(wrapped.absolute())
            elif not isinstance(wrapped, str):
                raise ValueError(
                    "wrapped must be a class or a str representing either the name of a "
                    "Docker image if executing the app as a Docker container, or the "
                    "name of the executable to run if extending the Docker image , not "
                    f"{wrapped!r}"
                )
            klass = None
            if wrapped.startswith("/"):
                executable = wrapped
                image_tag = None
            else:

                if wrapped.startswith("docker://"):
                    image_tag = wrapped.split("docker://")[-1]
                else:
                    logger.info(
                        "Assuming that the wrapped string, '%s' is a docker image tag. ",
                        wrapped,
                    )
                    image_tag = wrapped
                if "::" in image_tag:
                    image_tag, executable = image_tag.split("::")
                else:
                    executable = None  # entrypoint of the container
            if name is None:
                if image_tag:
                    class_name = image_tag.split("/")[-1].split(":")[0]
                else:
                    class_name = Path(executable).name.split(".")[0]
                    if class_name[0].isdigit():
                        class_name = DIGIT_TO_WORD[class_name[0]] + class_name[1:]
                class_name = re.sub(r"[^a-zA-Z0-9]", "_", class_name)

            parsed_inputs = (
                inputs if isinstance(inputs, dict) else {i.name: i for i in inputs}
            )
            parsed_outputs = (
                outputs if isinstance(outputs, dict) else {o.name: o for o in outputs}
            )

            # Add in fields from base classes
            parsed_inputs.update({n: getattr(Task, n) for n in Task.BASE_ATTRS})
            parsed_outputs.update({n: getattr(Outputs, n) for n in Outputs.BASE_ATTRS})

            parsed_inputs, parsed_outputs = ensure_field_objects(
                arg_type=arg,
                out_type=out,
                inputs=parsed_inputs,
                outputs=parsed_outputs,
                input_helps={},
                output_helps={},
            )
        if clashing := set(parsed_inputs) & set(["exectuable", "image_tag"]):
            raise ValueError(f"{list(clashing)} are reserved input names")

        parsed_inputs["executable"] = arg(
            name="executable", type=str, default=executable, path="not/used"
        )
        parsed_inputs["image_tag"] = arg(
            name="image_tag", type=str | None, default=image_tag, path="not/used"
        )

        defn = build_task_class(
            Task,
            Outputs,
            parsed_inputs,
            parsed_outputs,
            name=class_name,
            klass=klass,
            bases=bases,
            outputs_bases=outputs_bases,
            xor=xor,
        )

        return defn

    if wrapped is not None:
        if not isinstance(wrapped, (str, Path, type)):
            raise ValueError(f"wrapped must be a class or a str, not {wrapped!r}")
        return make(wrapped)
    return make


DIGIT_TO_WORD = {
    "0": "zero",
    "1": "one",
    "2": "two",
    "3": "three",
    "4": "four",
    "5": "five",
    "6": "six",
    "7": "seven",
    "8": "eight",
    "9": "nine",
}


def get_docker_entrypoint(image_tag: str) -> str:
    """Pulls a given Docker image tag and inspects the image to get its
    entrypoint/cmd

    IMAGE_TAG is the tag of the Docker image to inspect"""
    dc = docker.from_env()

    dc.images.pull(image_tag)

    image_attrs = dc.api.inspect_image(image_tag)["Config"]

    executable = image_attrs["Entrypoint"]
    if executable is None:
        executable = image_attrs["Cmd"]

    return executable



from pathlib import Path
from fileformats.generic import Directory
from pydra.compose import shell


@shell.define
class BidsApp(shell.Task["BidsApp.Outputs"]):

    executable = ""  # This should be changed to None once https://github.com/nipype/pydra/pull/822 is merged

    dataset_path: Directory = shell.arg(
        help="Path to BIDS dataset in the container",
        position=1,
        argstr="'{dataset_path}'",
    )

    output_path: Path = shell.arg(
        help="Directory where outputs will be written in the container",
        position=2,
        argstr="'{output_path}'",
    )

    analysis_level: str = shell.arg(
        help="The analysis level the app will be run at",
        position=3,
        argstr="",
        default="participant",
        allowed_values=["participant", "group"],
    )

    participant_label: str | None = shell.arg(
        help="The IDs to include in the analysis",
        argstr="--participant-label ",
        default=None,
        position=4,
    )

    flags: str | None = shell.arg(
        help="Additional flags to pass to the app",
        argstr="",
        default=None,
        position=-1,
    )

    work_dir: Path | None = shell.arg(
        help="Directory where the nipype temporary working directories will be stored",
        argstr="--work-dir '{work_dir}'",
        default=None,
    )

    class Outputs(shell.Outputs):

        pass
